/*-
 * Copyright (c) 2006 Allan Saddi <allan@saddi.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: ajp-glue.c 2260 2006-12-15 21:58:44Z asaddi $
 */

#include <Python.h>

#include <assert.h>

#include "ajp.h"
#include "wsgi-int.h"

/* Populate WSGI environ dictionary */
int
wsgiPopulateEnviron(Request *self)
{
  AJPRequest *req = ((AJPContext *)self->context)->activeReq;
  PyObject *environ = self->environ;
  PyObject *val;
  int ret, i, j, k;
  const char *pathInfo;
  char buf[256];

  if (wsgiPutEnv(self, "REQUEST_METHOD", req->method))
    return -1;
  if (wsgiPutEnv(self, "SERVER_PROTOCOL", req->protocol))
    return -1;
  if (wsgiPutEnv(self, "REQUEST_URI", req->uri))
    return -1;
  if (wsgiPutEnv(self, "REMOTE_ADDR", req->remoteAddr))
    return -1;
  if (wsgiPutEnv(self, "REMOTE_HOST", req->remoteHost))
    return -1;
  if (wsgiPutEnv(self, "SERVER_NAME", req->serverName))
    return -1;
  snprintf (buf, sizeof(buf), "%d", req->serverPort);
  if (wsgiPutEnv(self, "SERVER_PORT", buf))
    return -1;

  /* HTTP headers */
  for (i = 0; i < req->numHeaders; i++) {
    const char *header = req->headers[2 * i];
    const char *value = req->headers[2 * i + 1];

    /* Copy/convert header name */
    strcpy(buf, "HTTP_");

    k = strlen(header);
    if (k > (sizeof(buf) - 6))
      k = sizeof(buf) - 6;
    for (j = 0; j < k; j++) {
      if (header[j] == '-')
	buf[5 + j] = '_';
      else
	buf[5 + j] = toupper(header[j]);
    }
    buf[5 + j] = '\0';

    if (!strcmp(buf, "HTTP_CONTENT_TYPE") ||
	!strcmp(buf, "HTTP_CONTENT_LENGTH")) {
      /* Strip HTTP_ */
      if (wsgiPutEnv(self, &buf[5], value))
	return -1;
    }
    else {
      if (wsgiPutEnv(self, buf, value))
	return -1;
    }
  }

  /* Attributes */
  for (i = 0; i < req->numAttributes; i++) {
    const char *name = req->attributes[2 * i];
    const char *value = req->attributes[2 * i + 1];

    if (wsgiPutEnv(self, name, value))
      return -1;
  }

  /* SSL_KEY_SIZE attribute is special */
  if (req->sslKeySize) {
    snprintf(buf, sizeof(buf), "%d", req->sslKeySize);
    if (wsgiPutEnv(self, "SSL_KEY_SIZE", buf))
      return -1;
  }

  /* wsgi.version */
  if ((val = Py_BuildValue("(ii)", 1, 0)) == NULL)
    return -1;
  ret = PyDict_SetItemString(environ, "wsgi.version", val);
  Py_DECREF(val);
  if (ret)
    return -1;

  /* wsgi.input */
  if (PyDict_SetItemString(environ, "wsgi.input", self->input))
    return -1;

  /* wsgi.errors */
  if (PyDict_SetItemString(environ, "wsgi.errors", self->pStderr))
    return -1;

  /* wsgi.multithread */
  if (PyDict_SetItemString(environ, "wsgi.multithread", Py_True))
    return -1;

  /* wsgi.multiprocess */
  if (PyDict_SetItemString(environ, "wsgi.multiprocess", Py_False))
    return -1;

  /* wsgi.run_once */
  if (PyDict_SetItemString(environ, "wsgi.run_once", Py_False))
    return -1;

  /* wsgi.url_scheme */
  if (wsgiPutEnv(self, "wsgi.url_scheme", req->isSSL ? "https" : "http"))
    return -1;

  /* SCRIPT_NAME/PATH_INFO */
  pathInfo = req->uri;
  if (self->scriptNameLen) {
    if (!strncmp(req->uri, self->scriptName, self->scriptNameLen))
      pathInfo = &pathInfo[self->scriptNameLen];
    else
      fprintf (stderr, "WARNING: SCRIPT_NAME does not match REQUEST_URI\n");
  }

  if (wsgiPutEnv(self, "PATH_INFO", pathInfo))
    return -1;

  if (wsgiPutEnv(self, "SCRIPT_NAME", self->scriptName))
    return -1;

  /* Ensure QUERY_STRING is set */
  if (!PyDict_GetItemString(self->environ, "QUERY_STRING")) {
    if (wsgiPutEnv(self, "QUERY_STRING", ""))
      return -1;
  }

  return 0;
}

static int
_wsgiMapError(int error)
{
   assert(error < 0);

   switch (error) {
   case AJP_MEMORY_ERROR:
     PyErr_SetString(PyExc_MemoryError, "AJP: Out of memory");
     break;
   case AJP_SOCKET_ERROR:
     PyErr_SetString(PyExc_IOError, "AJP: Socket read/write error");
     break;
   case AJP_BUFFER_ERROR:
     PyErr_SetString(PyExc_IndexError, "AJP: Buffer overrun (protocol error)");
     break;
   case AJP_PROTOCOL_ERROR:
     PyErr_SetString(PyExc_ValueError, "AJP: Protocol error");
     break;
   default:
     PyErr_SetString(PyExc_RuntimeError, "AJP: Unknown error");
     break;
   }
   return -1;
}

int
wsgiSendHeaders(void *ctxt, int statusCode, const char *statusMsg,
		int headerCount, const char *headers[])
{
  int ret;

  Py_BEGIN_ALLOW_THREADS
  ret = ajpSendHeaders(ctxt, statusCode, statusMsg, headerCount, headers);
  Py_END_ALLOW_THREADS

  if (ret) return _wsgiMapError(ret);

  return 0;
}

int
wsgiSendBody(void *ctxt, const uint8_t *data, size_t len)
{
  int pos = 0, size, ret = AJP_OK;

  Py_BEGIN_ALLOW_THREADS
  while (pos < len && ret == AJP_OK) {
    size = len - pos;
    if (size > AJP_MAX_BODY_CHUNK_SIZE)
      size = AJP_MAX_BODY_CHUNK_SIZE;
    ret = ajpSendBodyChunk(ctxt, &data[pos], size);
    pos += size;
  }
  Py_END_ALLOW_THREADS

  if (ret) return _wsgiMapError(ret);

  return 0;
}

int
wsgiGetBody(void *ctxt)
{
  int ret;

  Py_BEGIN_ALLOW_THREADS
  ret = ajpGetBodyChunk(ctxt);
  Py_END_ALLOW_THREADS

  if (ret) return _wsgiMapError(ret);

  return 0;
}

int
wsgiPrimeInput(void *ctxt)
{
  int ret;

  if (((AJPContext *)ctxt)->activeReq->contentLength > 0) {
    Py_BEGIN_ALLOW_THREADS
    ret = ajpProcessInput(ctxt);
    Py_END_ALLOW_THREADS

    if (ret) return _wsgiMapError(ret);
  }
  return 0;
}

void
wsgiSetRequestData(void *ctxt, void *data)
{
  ((AJPContext *)ctxt)->activeReq->data = data;
}

void *
wsgiGetRequestData(void *ctxt)
{
  return ((AJPContext *)ctxt)->activeReq->data;
}

int
wsgiGetContentLength(void *ctxt)
{
  return ((AJPContext *)ctxt)->activeReq->contentLength;
}
