/*-
 * Copyright (c) 2006 Allan Saddi <allan@saddi.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: main.c 2266 2006-12-16 19:55:58Z asaddi $
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <libgen.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <limits.h>
#include <netdb.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pthread.h>

#include "ajp.h"
#include "wsgi.h"
#include "threadpool.h"
#include "version.h"

#define MAX_SOCKETS 10

static char *progName;
static int verbosity = 0;

static int socks[MAX_SOCKETS];
static int sockCount, nfds;

struct clientData {
  int sock;
  struct sockaddr_in addr;
};

static void
myperror(const char *subModule)
{
  if (subModule == NULL)
    fprintf(stderr, "%s: %s\n", progName, strerror(errno));
  else
    fprintf(stderr, "%s: %s: %s\n", progName, subModule, strerror(errno));
}

static void
clientHandler(void *clientData)
{
  struct clientData *cd = clientData;
  AJPContext *ctxt;

  if (verbosity)
    fprintf(stderr, "%s: Connection opened (%s:%d)\n",
	    progName,
	    inet_ntoa(cd->addr.sin_addr),
	    (int)ntohs(cd->addr.sin_port));
  if ((ctxt = ajpCreateContext(cd->sock,
			       (ajp_handler_t)wsgiHandler,
			       (ajp_body_handler_t)wsgiBodyHandler))) {
    /* Main loop */
    while (ajpProcessInput(ctxt) == AJP_OK);

    ajpDestroyContext(ctxt);
  }
  if (verbosity)
    fprintf(stderr, "%s: Connection closed (%s:%d)\n",
	    progName,
	    inet_ntoa(cd->addr.sin_addr),
	    (int)ntohs(cd->addr.sin_port));

  free(cd);
}

static int
createSocket(const char *ifname, int port)
{
  struct addrinfo hints, *res, *res0;
  int error;
  int s;
  int on = 1;
  char portBuf[10];

  snprintf(portBuf, sizeof(portBuf), "%d", port);

  memset(&hints, 0, sizeof(hints));
  hints.ai_family = PF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;
  error = getaddrinfo(ifname, portBuf, &hints, &res0);
  if (error) {
    fprintf(stderr, "%s: getaddrinfo: %s\n", progName,
	    gai_strerror(error));
    return -1;
  }

  sockCount = 0;
  nfds = -1;
  for (res = res0; res; res = res->ai_next) {
    if ((s = socket(res->ai_family, res->ai_socktype,
		    res->ai_protocol)) == -1) {
      myperror("socket");
      continue;
    }

    if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on))) {
      myperror("setsockopt");
      close(s);
      continue;
    }

    if (bind(s, res->ai_addr, res->ai_addrlen)) {
      myperror("bind");
      close(s);
      continue;
    }

    if (listen(s, SOMAXCONN)) {
      myperror("listen");
      close(s);
      continue;
    }

    socks[sockCount++] = s;

    if (s > nfds)
      nfds = s;

    if (sockCount == MAX_SOCKETS) {
      fprintf (stderr, "%s: Maximum number of sockets bound\n",
	       progName);
      break;
    }
  }
  freeaddrinfo(res0);

  nfds++;

  return sockCount ? 0 : -1;
}

static void
die(int foo)
{
  if (verbosity)
    fprintf(stderr, "%s: Exiting...\n", progName);
  exit(0);
}

static void
reload(int foo)
{
  if (verbosity)
    fprintf(stderr, "%s: Exiting for reload...\n", progName);
  exit(42);
}

static void
usage(void)
{
  int len, i;
  char *filler, *c;

  len = strlen(progName);
  if ((filler = malloc(len + 1)) == NULL) {
    myperror("malloc");
    exit(1);
  }
  for (i = 0, c = filler; i < len; i++)
    *(c++) = ' ';
  *c = '\0';

  fprintf(stderr,
	  "Usage: %s [-V] [-v] [-B] [-l <logFile>]\n"
	  "       %s [-h <ifname>] [-p <port>]\n"
	  "       %s [-M <maxThreads>] [-m <minThreads>]\n"
	  "       %s <moduleName> <appName> [<scriptName>]\n",
	  progName, filler, filler, filler);
  exit(1);
}

int
main(int argc, char *argv[])
{
  int ch;
  char *end;
  char *scriptName = "", *ifname = "localhost";
  int port = 8009;
  int daemonize = 0;
  char *logName = NULL;
  int minThreads = 10, maxThreads = INT_MAX;

  if ((progName = strdup(basename(argv[0]))) == NULL) {
    perror(NULL);
    exit(1);
  }

  while ((ch = getopt(argc, argv, "BM:Vl:m:h:p:v")) != -1) {
    switch (ch) {
    case 'B':
      daemonize++;
      break;
    case 'M':
      maxThreads = strtol(optarg, &end, 10);
      if (*end || maxThreads < 0) {
	fprintf(stderr, "%s: Bad maxThreads argument\n", progName);
	exit(1);
      }
      break;
    case 'V':
      fprintf (stderr, "%s %s\n", progName, VERSION_STRING);
      exit(1);
      break;
    case 'l':
      logName = optarg;
      break;
    case 'm':
      minThreads = strtol(optarg, &end, 10);
      if (*end || minThreads < 0) {
	fprintf(stderr, "%s: Bad minThreads argument\n", progName);
	exit(1);
      }
      break;
    case 'h':
      ifname = optarg;
      break;
    case 'p':
      port = strtol(optarg, &end, 10);
      if (*end || port < 1 || port > 65535) {
	fprintf (stderr, "%s: Bad port argument\n", progName);
	exit(1);
      }
      break;
    case 'v':
      verbosity++;
      break;
    default:
      usage();
    }
  }
  argc -= optind;
  argv += optind;

  if (maxThreads < minThreads) {
    fprintf(stderr, "%s: maxThreads must be >= minThreads\n", progName);
    exit(1);
  }

  if (argc < 2 || argc > 3)
    usage();

  if (argc > 2) {
    scriptName = argv[2];
    if (scriptName[0]) {
      if (scriptName[0] != '/') {
	fprintf(stderr, "%s: scriptName must start with '/'\n", progName);
	exit(1);
      }
      if (scriptName[strlen(scriptName) - 1] == '/') {
	fprintf(stderr, "%s: scriptName must not end with '/'\n", progName);
	exit(1);
      }
    }
  }

  if (createSocket(ifname, port))
    exit(2);

  if (logName != NULL) {
    int log;

    if ((log = open(logName, O_CREAT|O_APPEND|O_WRONLY, 0666)) == -1) {
      myperror(logName);
      exit(3);
    }

    dup2(log, STDERR_FILENO);

    if (log != STDERR_FILENO)
      close(log);
  }

  if (daemonize) {
    /* Do this ourselves so we have more control */
    int fd = open("/dev/null", O_RDWR);
    dup2(fd, STDIN_FILENO);
    dup2(fd, STDOUT_FILENO);
    if (logName == NULL)
      dup2(fd, STDERR_FILENO);
    if (fd > STDERR_FILENO)
      close(fd);
  }

  if (wsgiInit(argv[0], argv[1], scriptName, progName))
    exit(4);

  if (daemonize) {
    if (daemon (1, 1)) {
      myperror("daemon");
      exit(5);
    }
  }
  else
    signal(SIGHUP, reload);

  signal(SIGINT, die);
  signal(SIGTERM, die);

  if (threadpoolInit(minThreads, maxThreads)) {
    myperror("threadpoolInit");
    exit(6);
  }

  /* Main loop */
  for(;;) {
    fd_set readfds;
    int i;

    FD_ZERO(&readfds);
    for (i = 0; i < sockCount; i++)
      FD_SET(socks[i], &readfds);

    /* Wait for activity */
    if (select(nfds, &readfds, NULL, NULL,  NULL) == -1) {
      myperror("select");
      exit(7);
    }

    /* Scan for active sockets */
    for (i = 0; i < sockCount; i++) {
      if (FD_ISSET(socks[i], &readfds)) {
	struct clientData *cd;
	socklen_t addrLen = sizeof(cd->addr);

	if ((cd = malloc(sizeof(*cd))) == NULL) {
	  myperror("malloc");
	  exit(7);
	}

	if ((cd->sock = accept(socks[i], (struct sockaddr *)&cd->addr,
			       &addrLen)) != -1) {
	  if (threadpoolRun(clientHandler, cd)) {
	    myperror("threadpoolRun");
	    close(cd->sock);
	    free(cd);
	  }
	}
	else {
	  myperror("accept");
	  free(cd);
	}
      }
    }
  }
}
