/*-
 * Copyright (c) 2006 Allan Saddi <allan@saddi.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: threadpool.c 2264 2006-12-16 18:41:04Z asaddi $
 */

#include <errno.h>
#include <stdlib.h>

#include <pthread.h>

#include "threadpool.h"

typedef struct _QueuedJob {
  struct _QueuedJob *next;
  void (*job)(void *);
  void *arg;
} QueuedJob;

static pthread_mutex_t poolMutex;
static pthread_cond_t poolCond;
static QueuedJob *poolHead, *poolTail;
static int minSize, maxSize, poolSize, busyThreads;

static void *
jobRunner(void *arg)
{
  QueuedJob *job = arg;

  pthread_mutex_lock(&poolMutex);

  /* Main loop */
  for (;;) {
    if (job == NULL) {
      /* Wait for a job */
      while (poolHead == NULL)
	pthread_cond_wait(&poolCond, &poolMutex);

      /* Dequeue job */
      job = poolHead;
      poolHead = poolHead->next;

      if (poolTail == job) /* Last job? */
	poolTail = NULL;
    }

    busyThreads++;

    pthread_mutex_unlock(&poolMutex);

    /* Run the job */
    job->job(job->arg);

    free(job);
    job = NULL;

    pthread_mutex_lock(&poolMutex);

    busyThreads--;

    if (poolSize > minSize)
      break; /* NB: still locked */
  }

  poolSize--;

  /* Die off */
  pthread_mutex_unlock(&poolMutex);

  return NULL;
}

int
threadpoolInit(int _minSize, int _maxSize)
{
  int i;
  pthread_t t;

  if (pthread_mutex_init(&poolMutex, NULL))
    return -1;

  if (pthread_cond_init(&poolCond, NULL))
    return -1;

  poolHead = NULL;
  poolTail = NULL;

  minSize = poolSize = _minSize;
  maxSize = _maxSize;
  busyThreads = 0;

  pthread_mutex_lock(&poolMutex);

  for (i = 0; i < minSize; i++) {
    if (pthread_create(&t, NULL, jobRunner, NULL))
      return -1;
    else
      pthread_detach(t);
  }

  pthread_mutex_unlock(&poolMutex);

  return 0;
}

int
threadpoolRun(void (*_job)(void *), void *arg)
{
  int result = -1;
  QueuedJob *job;

  if ((job = malloc(sizeof(*job))) == NULL)
    return -1;

  job->next = NULL;
  job->job = _job;
  job->arg = arg;

  pthread_mutex_lock(&poolMutex);

  /* Anymore threads? */
  if (busyThreads == poolSize) {
    if (poolSize == maxSize) {
      /* No more */
      errno = EAGAIN;
      goto out;
    }
    else {
      pthread_t t;

      /* Start new thread */
      if (pthread_create(&t, NULL, jobRunner, job)) {
	goto out;
      }
      else {
	pthread_detach(t);
	poolSize++;
      }
    }
  }
  else {
    /* Queue it up */
    if (poolHead == NULL) {
      /* First job */
      poolHead = job;
      poolTail = job;
    }
    else {
      poolTail->next = job;
      poolTail = job;
    }
    pthread_cond_signal(&poolCond);
  }

  result = 0;

 out:
  pthread_mutex_unlock(&poolMutex);
  return result;
}
